import logging
import re
import pubchempy as pcp
import numpy as np
from matchms.utils import is_valid_inchikey


def pubchem_metadata_lookup(spectrum_in, name_search_depth=10, match_precursor_mz=False,
                            formula_search=False,
                            min_formula_length=6, formula_search_depth=25, verbose=1):
    """

    Parameters
    ----------
    spectrum_in
        Matchms type spectrum as input.
    name_search_depth: int
        How many of the most relevant name matches to explore deeper. Default = 10.

    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()
    if is_valid_inchikey(spectrum.get("inchikey")):
        return spectrum

    def _plausible_name(compound_name):
        return (isinstance(compound_name, str) and len(compound_name) > 4)

    compound_name = spectrum.get("compound_name")
    if not _plausible_name(compound_name):
        return spectrum

    # Start pubchem search
    inchi = spectrum.get("inchi")
    parent_mass = spectrum.get("parent_mass")
    if isinstance(parent_mass, np.ndarray):
        parent_mass = parent_mass[0]
    formula = spectrum.get("formula")

    # 1) Search for matching compound name
    results_pubchem = pubchem_name_search(compound_name, name_search_depth=name_search_depth,
                                          verbose=verbose)

    if len(results_pubchem) > 0:

        # 1a) Search for matching inchi
        if likely_has_inchi(inchi):
            inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_inchi_match(results_pubchem, inchi,
                                                                                       verbose=verbose)
        # 1b) Search for matching parent mass
        if not likely_has_inchi(inchi) or inchikey_pubchem is None:
            inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_mass_match(results_pubchem,
                                                                                      parent_mass,
                                                                                      given_mass="parent mass",
                                                                                      verbose=verbose)

        # 1c) Search for matching precursor mass (optional)
        if match_precursor_mz and inchikey_pubchem is None:
            precursor_mz = spectrum.get("precursor_mz")
            inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_mass_match(results_pubchem,
                                                                                      precursor_mz,
                                                                                      given_mass="precursor mass",
                                                                                      verbose=verbose)


        if inchikey_pubchem is not None and inchi_pubchem is not None:
            logging.info("Matching compound name: %s", compound_name)
            if verbose >= 1:
                print(f"Matching compound name: {compound_name}")
            spectrum.set("inchikey", inchikey_pubchem)
            spectrum.set("inchi", inchi_pubchem)
            spectrum.set("smiles", smiles_pubchem)
            return spectrum

        if verbose >= 2:
            print(f"No matches found for compound name: {compound_name}")

    # 2) Search for matching formula
    if formula_search and formula and len(formula) >= min_formula_length:
        results_pubchem = pubchem_formula_search(formula, formula_search_depth=formula_search_depth,
                                                 verbose=verbose)

        if len(results_pubchem) > 0:

            # 2a) Search for matching inchi
            if likely_has_inchi(inchi):
                inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_inchi_match(results_pubchem, inchi,
                                                                                           verbose=verbose)
            # 2b) Search for matching parent mass
            if inchikey_pubchem is None:
                inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_mass_match(results_pubchem,
                                                                                          parent_mass,
                                                                                          given_mass="parent mass",
                                                                                          verbose=verbose)
            # 2c) Search for matching precursor mass (optional)
            if match_precursor_mz and inchikey_pubchem is None:
                precursor_mz = spectrum.get("precursor_mz")
                inchi_pubchem, inchikey_pubchem, smiles_pubchem = find_pubchem_mass_match(results_pubchem,
                                                                                          precursor_mz,
                                                                                          given_mass="precursor mass",
                                                                                          verbose=verbose)
            if inchikey_pubchem is not None and inchi_pubchem is not None:
                logging.info("Matching formula: %s", formula)
                if verbose >= 1:
                    print(f"Matching formula: {formula}")
                spectrum.set("inchikey", inchikey_pubchem)
                spectrum.set("inchi", inchi_pubchem)
                spectrum.set("smiles", smiles_pubchem)
                return spectrum

            if verbose >= 2:
                print(f"No matches found for formula: {formula}")

    return spectrum


def likely_has_inchi(inchi):
    """Quick test to avoid excess in-depth testing"""
    if inchi is None:
        return False
    inchi = inchi.strip('"')
    regexp = r"(InChI=1|1)(S\/|\/)[0-9, A-Z, a-z,\.]{2,}\/(c|h)[0-9]"
    if not re.search(regexp, inchi):
        return False
    return True


def likely_inchi_match(inchi_1, inchi_2, min_agreement=3):
    """Try to match defective inchi to non-defective ones.

    Compares inchi parts seperately. Match is found if at least the first
    'min_agreement' parts are a good enough match.
    The main 'defects' this method accounts for are missing '-' in the inchi.
    In addition, differences between '-', '+', and '?'will be ignored.

    Parameters
    ----------
    inchi_1: str
        inchi of molecule.
    inchi_2: str
        inchi of molecule.
    min_agreement: int
        Minimum number of first parts that MUST be a match between both input
        inchi to finally consider it a match. Default is min_agreement=3.
    """
    if min_agreement < 2:
        print("Warning! 'min_agreement' < 2 has no discriminative power. Should be => 2.")
    if min_agreement == 2:
        print("Warning! 'min_agreement' == 2 has little discriminative power",
              "(only looking at structure formula. Better use > 2.")
    agreement = 0

    # Remove spaces and '"' to account for different notations.
    # Remove everything with little discriminative power.
    ignore_lst = ['"', ' ', '-', '+', '?']
    for ignore in ignore_lst:
        inchi_1 = inchi_1.replace(ignore, '')
        inchi_2 = inchi_2.replace(ignore, '')

    # Split inchi in parts.
    inchi_1_parts = inchi_1.split('/')
    inchi_2_parts = inchi_2.split('/')

    # Check if both inchi have sufficient parts (seperated by '/')
    if len(inchi_1_parts) >= min_agreement and len(
            inchi_2_parts) >= min_agreement:
        # Count how many parts agree well
        for i in range(min_agreement):
            agreement += (inchi_1_parts[i] == inchi_2_parts[i])

    return bool(agreement == min_agreement)


def likely_inchikey_match(inchikey_1, inchikey_2, min_agreement=1):
    """Try to match inchikeys.

    Compares inchikey parts seperately. Match is found if at least the first
    'min_agreement' parts are a good enough match.

    Parameters
    ----------
    inchikey_1: str
        inchikey of molecule.
    inchikey_2: str
        inchikey of molecule.
    min_agreement: int
        Minimum number of first parts that MUST be a match between both input
        inchikey to finally consider it a match. Default is min_agreement=1.
    """
    if min_agreement not in [1, 2, 3]:
        print("Warning! 'min_agreement' should be 1, 2, or 3.")
    agreement = 0

    # Harmonize strings
    inchikey_1 = inchikey_1.upper().replace('"', '').replace(' ', '')
    inchikey_2 = inchikey_2.upper().replace('"', '').replace(' ', '')

    # Split inchikey in parts.
    inchikey_1_parts = inchikey_1.split('-')
    inchikey_2_parts = inchikey_2.split('-')

    # Check if both inchikey have sufficient parts (seperated by '/')
    if len(inchikey_1_parts) >= min_agreement and len(
            inchikey_2_parts) >= min_agreement:
        # Count how many parts mostly agree
        for i in range(min_agreement):
            agreement += (inchikey_1_parts[i] == inchikey_2_parts[i])

    return agreement == min_agreement


def pubchem_name_search(compound_name: str, name_search_depth=10, verbose=1):
    """Search pubmed for compound name"""
    results_pubchem = pcp.get_compounds(compound_name,
                                        'name',
                                        listkey_count=name_search_depth)
    if verbose >=2:
        print("Found at least", len(results_pubchem),
              "compounds of that name on pubchem.")
    return results_pubchem


def pubchem_formula_search(compound_formula: str, formula_search_depth=25, verbose=1):
    """Search pubmed for compound formula"""
    sids_pubchem = pcp.get_sids(compound_formula,
                                'formula',
                                listkey_count=formula_search_depth)

    results_pubchem = []
    for sid in sids_pubchem:
        result = pcp.Compound.from_cid(sid['CID'])
        results_pubchem.append(result)

    if verbose >=2:
        print(f"Found at least {len(results_pubchem)} compounds of with formula: {compound_formula}.")
    return results_pubchem


def find_pubchem_inchi_match(results_pubchem,
                             inchi,
                             min_inchi_match=3,
                             verbose=1):
    """Searches pubmed matches for inchi match.
    Then check if inchi can be matched to (defective) input inchi.


    Outputs found inchi and found inchikey (will be None if none is found).

    Parameters
    ----------
    results_pubchem: List[dict]
        List of name search results from Pubchem.
    inchi: str
        Inchi (correct, or defective...). Set to None to ignore.
    min_inchi_match: int
        Minimum number of first parts that MUST be a match between both input
        inchi to finally consider it a match. Default is min_inchi_match=3.
    """

    inchi_pubchem = None
    inchikey_pubchem = None
    smiles_pubchem = None

    # Loop through first 'name_search_depth' results found on pubchem. Stop once first match is found.
    for result in results_pubchem:
        inchi_pubchem = '"' + result.inchi + '"'
        inchikey_pubchem = result.inchikey
        smiles_pubchem = result.isomeric_smiles
        if smiles_pubchem is None:
            smiles_pubchem = result.canonical_smiles

        match_inchi = likely_inchi_match(inchi, inchi_pubchem,
                                         min_agreement=min_inchi_match)

        if match_inchi:
            logging.info("Matching inchi: %s", inchi)
            if verbose >= 1:
                print(f"Found matching compound for inchi: {inchi} (Pubchem: {inchi_pubchem}")
            break

    if not match_inchi:
        inchi_pubchem = None
        inchikey_pubchem = None
        smiles_pubchem = None

        if verbose >= 2:
            print("No matches found for inchi", inchi, "\n")

    return inchi_pubchem, inchikey_pubchem, smiles_pubchem


def find_pubchem_mass_match(results_pubchem,
                            parent_mass,
                            mass_tolerance=2.0,
                            given_mass="parent mass",
                            verbose=1):
    """Searches pubmed matches for inchi match.
    Then check if inchi can be matched to (defective) input inchi.


    Outputs found inchi and found inchikey (will be None if none is found).

    Parameters
    ----------
    results_pubchem: List[dict]
        List of name search results from Pubchem.
    parent_mass: float
        Spectrum"s guessed parent mass.
    mass_tolerance: float
        Acceptable mass difference between query compound and pubchem result.
    given_mass
        String to specify the type of the given mass (e.g. "parent mass").
    """
    inchi_pubchem = None
    inchikey_pubchem = None
    smiles_pubchem = None

    for result in results_pubchem:
        inchi_pubchem = '"' + result.inchi + '"'
        inchikey_pubchem = result.inchikey
        smiles_pubchem = result.isomeric_smiles
        if smiles_pubchem is None:
            smiles_pubchem = result.canonical_smiles

        pubchem_mass = float(results_pubchem[0].exact_mass)
        match_mass = (np.abs(pubchem_mass - parent_mass) <= mass_tolerance)

        if match_mass:
            logging.info("Matching molecular weight %s vs parent mass of %s",
                         str(np.round(pubchem_mass,1)),
                         str(np.round(parent_mass,1)))
            if verbose >= 1:
                print(f"Matching molecular weight ({pubchem_mass:.1f} vs {given_mass} of {parent_mass:.1f})")
            break

    if not match_mass:
        inchi_pubchem = None
        inchikey_pubchem = None
        smiles_pubchem = None

        if verbose >= 2:
            print(f"No matches found for mass {parent_mass} Da")

    return inchi_pubchem, inchikey_pubchem, smiles_pubchem



"""Plotting functions for spec2vec"""
import numpy as np
from matchms.similarity.spectrum_similarity_functions import collect_peak_pairs
from matplotlib import pyplot as plt
from scipy import spatial
from rdkit import Chem
from rdkit.Chem import Draw
from IPython.display import SVG, display
from spec2vec import SpectrumDocument
from matchms.filtering import normalize_intensities
from matchms.filtering import select_by_mz
from matchms.filtering import select_by_relative_intensity


def plot_precentile(arr_ref, arr_sim, num_bins=1000, show_top_percentile=1.0,
                    ignore_diagonal=False):
    """ Plot top percentile (as specified by show_top_percentile) of best restults
    in arr_sim and compare against reference values in arr_ref.

    Args:
    -------
    arr_ref: numpy array
        Array of reference values to evaluate the quality of arr_sim.
    arr_sim: numpy array
        Array of similarity values to evaluate.
    num_bins: int
        Number of bins to divide data (default = 1000)
    show_top_percentile
        Choose which part to plot. Will plot the top 'show_top_percentile' part of
        all similarity values given in arr_sim. Default = 1.0
    """
    def _ignore_reference_nans(arr_ref, arr_sim):
        assert arr_ref.shape == arr_sim.shape, "Expected two arrays of identical shape."
        idx_not_nans = np.where(np.isnan(arr_ref) == False)
        arr_sim = arr_sim[idx_not_nans]
        arr_ref = arr_ref[idx_not_nans]
        return arr_ref, arr_sim

    if ignore_diagonal:
        np.fill_diagonal(arr_ref, np.nan)

    arr_ref, arr_sim = _ignore_reference_nans(arr_ref, arr_sim)

    start = int(arr_sim.shape[0] * show_top_percentile / 100)
    idx = np.argpartition(arr_sim, -start)
    starting_point = arr_sim[idx[-start]]
    if starting_point == 0:
        print("not enough datapoints != 0 above given top-precentile")

    # Remove all data below show_top_percentile
    low_as = np.where(arr_sim < starting_point)[0]

    length_selected = arr_sim.shape[0] - low_as.shape[0]  # start+1

    data = np.zeros((2, length_selected))
    data[0, :] = np.delete(arr_sim, low_as)
    data[1, :] = np.delete(arr_ref, low_as)
    data = data[:, np.lexsort((data[1, :], data[0, :]))]

    ref_score_cum = []

    for i in range(num_bins):
        low = int(i * length_selected / num_bins)
        # high = int((i+1) * length_selected/num_bins)
        ref_score_cum.append(np.mean(data[1, low:]))
    ref_score_cum = np.array(ref_score_cum)
    x_percentiles = (show_top_percentile / num_bins * (1 + np.arange(num_bins)))[::-1]

    fig, ax = plt.subplots(figsize=(6, 6))
    plt.plot(
        x_percentiles,
        ref_score_cum,
        color='black')
    plt.xticks(np.linspace(0, show_top_percentile, 5),
               ["{:.2f}%".format(x) for x in np.linspace(0, show_top_percentile, 5)])
    plt.xlabel("Top percentile of spectral similarity score g(s,s')")
    plt.ylabel("Mean molecular similarity (f(t,t') within that percentile)")

    return ref_score_cum


def plot_spectra_comparison(spectrum1_in, spectrum2_in,
                            model,
                            intensity_weighting_power=0.5,
                            num_decimals=2,
                            min_mz=5,
                            max_mz=500,
                            intensity_threshold=0.01,
                            method='cosine',
                            tolerance=0.005,
                            wordsim_cutoff=0.5,
                            circle_size=5,
                            circle_scaling='wordsim',
                            padding=10,
                            display_molecules=False,
                            figsize=(12, 12),
                            filename=None):
    """ In-depth visual comparison of spectral similarity scores,
    calculated based on cosine/mod.cosine and Spev2Vec.

    Parameters
    ----------
    method: str
        'cosine' or 'modcos' (modified cosine score)
    circle_scaling: str
        Scale circles based on 'wordsim' or 'peak_product'
    """

    def apply_filters(s):
        s = normalize_intensities(s)
        s = select_by_mz(s, mz_from=min_mz, mz_to=max_mz)
        s = select_by_relative_intensity(s, intensity_from=intensity_threshold)
        s.losses = None
        return s

    spectrum1 = apply_filters(spectrum1_in)
    spectrum2 = apply_filters(spectrum2_in)

    plt.style.use("seaborn-white")#('ggplot')
    plot_colors = ['darkcyan', 'purple']

    # Definitions for the axes
    left, width = 0.1, 0.6
    bottom, height = 0.1, 0.6
    spacing = 0.01

    rect_wordsim = [left, bottom, width, height]
    rect_specx = [left, bottom + height + spacing, width, 0.2]
    rect_specy = [left + width + spacing, bottom, 0.25, height]

    document1 = SpectrumDocument(spectrum1, n_decimals=num_decimals)
    document2 = SpectrumDocument(spectrum2, n_decimals=num_decimals)

    # Remove words/peaks that are not in dictionary
    select1 = np.asarray([i for i, word in enumerate(document1.words) if word in model.wv.vocab])
    select2 = np.asarray([i for i, word in enumerate(document2.words) if word in model.wv.vocab])
    peaks1 = np.asarray(spectrum1.peaks[:]).T
    peaks2 = np.asarray(spectrum2.peaks[:]).T
    peaks1 = peaks1[select1, :]
    peaks2 = peaks2[select2, :]
    min_peaks1 = np.min(peaks1[:, 0])
    min_peaks2 = np.min(peaks2[:, 0])
    max_peaks1 = np.max(peaks1[:, 0])
    max_peaks2 = np.max(peaks2[:, 0])
    possible_grid_points = np.arange(0, 2000, 50)
    grid_points1 = possible_grid_points[(possible_grid_points > min_peaks1 - padding) \
                                        & (possible_grid_points < max_peaks1 + padding)]
    grid_points2 = possible_grid_points[(possible_grid_points > min_peaks2 - padding) \
                                        & (possible_grid_points < max_peaks2 + padding)]

    word_vectors1 = model.wv[[document1.words[x] for x in select1]]
    word_vectors2 = model.wv[[document2.words[x] for x in select2]]

    csim_words = 1 - spatial.distance.cdist(word_vectors1, word_vectors2, 'cosine')
    csim_words[csim_words < wordsim_cutoff] = 0  # Remove values below cutoff
    print(np.min(csim_words), np.max(csim_words))

    # Plot spectra
    # -------------------------------------------------------------------------
    fig = plt.figure(figsize=figsize)
    # Word similariy plot (central)
    ax_wordsim = plt.axes(rect_wordsim)
    ax_wordsim.tick_params(direction='in', top=True, right=True)
    # Spectra plot (top)
    ax_specx = plt.axes(rect_specx)
    ax_specx.tick_params(direction='in', labelbottom=False)
    # Spectra plot 2 (right)
    ax_specy = plt.axes(rect_specy)
    ax_specy.tick_params(direction='in', labelleft=False)

    # Spec2Vec similarity plot:
    # -------------------------------------------------------------------------
    data_x = []
    data_y = []
    data_z = []
    data_peak_product = []
    for i in range(len(select1)):
        for j in range(len(select2)):
            data_x.append(peaks1[i, 0])
            data_y.append(peaks2[j, 0])
            data_z.append(csim_words[i, j])
            data_peak_product.append(peaks1[i, 1] * peaks2[j, 1])

    # Sort by word similarity
    data_x = np.array(data_x)
    data_y = np.array(data_y)
    data_z = np.array(data_z)
    data_peak_product = np.array(data_peak_product)
    idx = np.lexsort((data_x, data_y, data_z))

    cm = plt.cm.get_cmap('RdYlBu_r')  # 'YlOrRd') #'RdBu_r')

    # Plot word similarities
    if circle_scaling == 'peak_product':
        wordsimplot = ax_wordsim.scatter(data_x[idx],
                                         data_y[idx],
                                         s=100 * circle_size *
                                         (0.01 + data_peak_product[idx]**2),
                                         marker="o",
                                         c=data_z[idx],
                                         cmap=cm,
                                         alpha=0.6)
    elif circle_scaling == 'wordsim':
        wordsimplot = ax_wordsim.scatter(data_x[idx],
                                         data_y[idx],
                                         s=100 * circle_size *
                                         (0.01 + data_z[idx]**2),
                                         marker="o",
                                         c=data_z[idx],
                                         cmap=cm,
                                         alpha=0.6)

    # (Modified) Cosine similarity plot:
    # -------------------------------------------------------------------------
    if method == 'cosine':
        score_classical, used_matches = cosine_score(spectrum1, spectrum2, tolerance, modified_cosine=False)
    elif method == 'modcos':
        score_classical, used_matches = cosine_score(spectrum1, spectrum2, tolerance, modified_cosine=True)
    else:
        print("Given method unkown.")

    idx1, idx2, _ = zip(*used_matches)
    cosine_x = []
    cosine_y = []
    for i in range(len(idx1)):
        if idx1[i] in select1 and idx2[i] in select2:
            cosine_x.append(peaks1[idx1[i], 0])
            cosine_y.append(peaks2[idx2[i], 0])

    # Plot (mod.) cosine similarities
    ax_wordsim.scatter(cosine_x, cosine_y, s=100, c='black', marker=(5, 2))
    ax_wordsim.set_xlim(min_peaks1 - padding, max_peaks1 + padding)
    ax_wordsim.set_ylim(min_peaks2 - padding, max_peaks2 + padding)
    ax_wordsim.set_xlabel('spectrum 1 - fragment mz', fontsize=16)
    ax_wordsim.set_ylabel('spectrum 2 - fragment mz', fontsize=16)
    ax_wordsim.tick_params(labelsize=13)
    ax_wordsim.set_xticks(grid_points1)
    ax_wordsim.set_yticks(grid_points2)
    ax_wordsim.grid(True)

    # Plot spectra 1
    ax_specx.vlines(peaks1[:, 0], [0], peaks1[:, 1], color=plot_colors[0])
    ax_specx.plot(peaks1[:, 0], peaks1[:, 1], '.')  # Stem ends
    ax_specx.plot([peaks1[:, 0].max(), peaks1[:, 0].min()], [0, 0],
                  '--')  # Middle bar
    ax_specx.set_xlim(min_peaks1 - padding, max_peaks1 + padding)
    ax_specx.set_yticks([0,0.25,0.5,0.75,1])
    ax_specx.set_xticks(grid_points1)
    ax_specx.set_ylabel('peak intensity (relative)', fontsize=16)
    ax_specx.tick_params(labelsize=13)

    ax_specx.grid(True)

    # Plot spectra 2
    ax_specy.hlines(peaks2[:, 0], [0], peaks2[:, 1], color=plot_colors[1])
    ax_specy.plot(peaks2[:, 1], peaks2[:, 0], '.')  # Stem ends
    ax_specy.plot([0, 0], [peaks2[:, 0].min(), peaks2[:, 0].max()],
                  '--')  # Middle bar
    ax_specy.set_ylim(min_peaks2 - padding, max_peaks2 + padding)
    ax_specy.set_xticks([0,0.25,0.5,0.75,1])
    ax_specy.set_yticks(grid_points2)
    ax_specy.set_xlabel('peak intensity (relative)', fontsize=16)
    ax_specy.tick_params(labelsize=13)

    ax_specy.grid(True)

    fig.colorbar(wordsimplot, ax=ax_specy)
    if filename is not None:
        plt.savefig(filename)
    plt.show()

    # Plot molecules
    # -------------------------------------------------------------------------
    if display_molecules:
        smiles = [spectrum1.get("smiles"), spectrum2.get("smiles")]
        molecules = [Chem.MolFromSmiles(x) for x in smiles]
        display(Draw.MolsToGridImage(molecules, molsPerRow=2, subImgSize=(400, 400)))


# def scour_svg_cleaning(target, source, env=[]):
#     """ Use scour to clean an svg file.

#     """
#     options = scour.generateDefaultOptions()

#     # override defaults for max cleansing
#     options.enable_viewboxing = True
#     options.strip_comments = True
#     options.strip_ids = True
#     options.remove_metadata = True
#     options.indent_type = None
#     options.shorten_ids = True

#     if 'SCOUR_OPTIONS' in env:
#         options.__dict__.update(env['SCOUR_OPTIONS'])

#     instream = open(source, 'rb')
#     outstream = open(target, 'wb')

#     scour.start(options, instream, outstream)


# def plot_molecules(smiles_lst, filename=None):
#     """ Plot molecule from smile(s).
#     Uses Scour to clean rdkit svg.

#     filename: str
#         If filename is given, molecules will be saved to filename.
#     """
#     if not isinstance(smiles_lst, list):
#         smiles_lst = [smiles_lst]
#     for i, smiles in enumerate(smiles_lst):
#         temp_file = "draw_mols_temp.svg"
#         mol = Chem.MolFromSmiles(smiles)
#         Draw.MolToFile(mol, temp_file)

#         # Clean svg using scour
#         if filename is not None:
#             file = filename.split('.svg')[0] + str(i) + '.svg'
#         else:
#             file = "draw_mols_temp_corr.svg"
#         scour(file, temp_file, [])

#         # Display cleaned svg
#         display(SVG(filename=temp_file))


def cosine_score(spectrum1, spectrum2, tolerance, modified_cosine=False):
    """

    Parameters
    ----------
    spectrum1 : TYPE
        DESCRIPTION.
    spectrum2 : TYPE
        DESCRIPTION.
    tolerance : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    def get_peaks_arrays():
        """Get peaks mz and intensities as numpy array."""
        spec1 = np.vstack((spectrum1.peaks.mz, spectrum1.peaks.intensities)).T
        spec2 = np.vstack((spectrum2.peaks.mz, spectrum2.peaks.intensities)).T
        assert max(spec1[:, 1]) <= 1, ("Input spectrum1 is not normalized. ",
                                       "Apply 'normalize_intensities' filter first.")
        assert max(spec2[:, 1]) <= 1, ("Input spectrum2 is not normalized. ",
                                       "Apply 'normalize_intensities' filter first.")
        return spec1, spec2

    def get_matching_pairs():
        """Get pairs of peaks that match within the given tolerance."""
        zero_pairs = collect_peak_pairs(spec1, spec2, tolerance, shift=0.0)
        if modified_cosine:
            message = "Precursor_mz missing. Apply 'add_precursor_mz' filter first."
            assert spectrum1.get("precursor_mz") and spectrum2.get("precursor_mz"), message
            mass_shift = spectrum1.get("precursor_mz") - spectrum2.get("precursor_mz")
            nonzero_pairs = collect_peak_pairs(spec1, spec2, tolerance, shift=mass_shift)
            unsorted_matching_pairs = zero_pairs + nonzero_pairs
        else:
            unsorted_matching_pairs = zero_pairs
        return sorted(unsorted_matching_pairs, key=lambda x: x[2], reverse=True)

    def calc_score():
        """Calculate cosine similarity score."""
        used1 = set()
        used2 = set()
        score = 0.0
        used_matches = []
        for match in matching_pairs:
            if not match[0] in used1 and not match[1] in used2:
                score += match[2]
                used1.add(match[0])  # Every peak can only be paired once
                used2.add(match[1])  # Every peak can only be paired once
                used_matches.append(match)
        # Normalize score:
        score = score/max(np.sum(spec1[:, 1]**2), np.sum(spec2[:, 1]**2))
        return score, used_matches

    spec1, spec2 = get_peaks_arrays()
    matching_pairs = get_matching_pairs()
    return calc_score()



from typing import List
import numpy as np
import pandas as pd
from tqdm import tqdm
from gensim.models.basemodel import BaseTopicModel
from matchms.similarity import CosineGreedy, ModifiedCosine, PrecursorMzMatch
from spec2vec import SpectrumDocument
from spec2vec import Spec2Vec


def library_matching(documents_query: List[SpectrumDocument],
                     documents_library: List[SpectrumDocument],
                     model: BaseTopicModel,
                     presearch_based_on: List[str] = ["precursor_mz", "spec2vec-top10"],
                     ignore_non_annotated: bool = True,
                     include_scores=["spec2vec", "cosine", "modcosine"],
                     intensity_weighting_power: float = 0.5,
                     allowed_missing_percentage: float = 0,
                     cosine_tol: float = 0.005,
                     min_matches: int = 6,
                     mass_tolerance: float = 2.0,
                     mass_tolerance_type: str = "ppm"):
    """Selecting potential spectra matches with spectra library.

    Suitable candidates will be selected by 1) top_n Spec2Vec similarity, and 2)
    same precursor mass (within given mz_ppm tolerance(s)).
    For later matching routines, additional scores (cosine, modified cosine)
    are added as well.

    Args:
    --------
    documents_query:
        List containing all spectrum documents that should be queried against the library.
    documents_library:
        List containing all library spectrum documents.
    model:
        Pretrained word2Vec model.
    presearch_based_on:
        List with strings to specify which measures to use for the presearch.
        This can include 'precursor_mz', 'spec2vec-topX',
    ignore_non_annotated: bool, optional
        If True, only annotated spectra will be considered for matching.
        Default = True.
    cosine_tol: float, optional
        Set tolerance for the cosine and modified cosine score. Default = 0.005
    mass_tolerance
        Specify tolerance for a mass match.
    mass_toleramce_type
        Chose between "ppm" (relative) and "Dalton" (absolute) tolerance type.
    """

    # Initializations
    found_matches = []
    m_mass_matches = None
    m_spec2vec_similarities = None
    m_modcos_similarities = None

    def get_metadata(documents):
        metadata = []
        for doc in documents:
            metadata.append(doc._obj.get("smiles"))
        return metadata

    library_spectra_metadata = get_metadata(documents_library)
    if ignore_non_annotated:
        # Get array of all ids for spectra with smiles
        library_ids = np.asarray([i for i, x in enumerate(library_spectra_metadata) if x])
    else:
        library_ids = np.arange(len(documents_library))

    allowed_presearch_type = ["precursor_mz", "spec2vec-top", "modcos-top"]
    msg = "Presearch must include one of: " + ", ".join(allowed_presearch_type)
    assert np.any([(x in y) for x in allowed_presearch_type for y in presearch_based_on]), msg

    # 1. Search for top-n Spec2Vec matches ------------------------------------
    if np.any(["spec2vec" in x for x in presearch_based_on]):
        top_n = int([x.split("top")[1] for x in presearch_based_on if "spec2vec" in x][0])
        print(f"Pre-selection includes spec2vec top {top_n}.")
        spec2vec = Spec2Vec(model=model, intensity_weighting_power=intensity_weighting_power,
                            allowed_missing_percentage=allowed_missing_percentage,
                            progress_bar=True)
        m_spec2vec_similarities = spec2vec.matrix([documents_library[i] for i in library_ids],
                                                  documents_query)

        # Select top_n similarity values:
        selection_spec2vec = np.argpartition(m_spec2vec_similarities, -top_n, axis=0)[-top_n:, :]
    else:
        selection_spec2vec = np.empty((0, len(documents_query)), dtype="int")

    # 2. Search for precursor_mz based matches ---------------------------------
    if "precursor_mz" in presearch_based_on:
        print(f"Pre-selection includes mass matches within {mass_tolerance} {mass_tolerance_type}.")
        mass_matching = PrecursorMzMatch(tolerance=mass_tolerance,
                                         tolerance_type=mass_tolerance_type)
        m_mass_matches = mass_matching.matrix([documents_library[i]._obj for i in library_ids],
                                              [x._obj for x in documents_query])
        selection_massmatch = []
        for i in range(len(documents_query)):
            selection_massmatch.append(np.where(m_mass_matches[:, i] == 1)[0])
    else:
        selection_massmatch = np.empty((len(documents_query), 0), dtype="int")

    # 3. Search for top-n modified cosine matches ------------------------------------
    if np.any(["modcos" in x for x in presearch_based_on]):
        top_n = int([x.split("top")[1] for x in presearch_based_on if "modcos" in x][0])
        print(f"Pre-selection includes modified cosine top {top_n}.")
        modcos = ModifiedCosine(tolerance=cosine_tol)

        n_rows = len(library_ids)
        n_cols = len(documents_query)
        m_modcos_similarities = np.zeros([n_rows, n_cols], dtype=np.float64)
        m_modcos_matches = np.zeros([n_rows, n_cols], dtype=np.float64)
        for i_ref, reference in enumerate(tqdm([documents_library[i]._obj for i in library_ids])):
            for i_query, query in enumerate([x._obj for x in documents_query]):
                score = modcos.pair(reference, query)
                m_modcos_similarities[i_ref][i_query] = score[0]
                m_modcos_matches[i_ref][i_query] = score[1]

        # Select top_n similarity values:
        m_modcos_selected = m_modcos_similarities.copy()
        m_modcos_selected[m_modcos_matches < min_matches] = 0
        selection_modcos = np.argpartition(m_modcos_selected, -top_n, axis=0)[-top_n:, :]
    else:
        selection_modcos = np.empty((0, len(documents_query)), dtype="int")

    # 4. Combine found matches ------------------------------------------------
    if "cosine" in include_scores:
        print("Calculate cosine score for selected candidates.")
    if "modcosine" in include_scores:
        print("Calculate modified cosine score for selected candidates.")

    for i in tqdm(range(len(documents_query))):
        s2v_top_ids = selection_spec2vec[:, i]
        mass_match_ids = selection_massmatch[i]
        modcos_ids = selection_modcos[:, i]

        all_match_ids = np.unique(np.concatenate((s2v_top_ids, mass_match_ids, modcos_ids)))

        if len(all_match_ids) > 0:
            if "cosine" in include_scores:
                # Get cosine score for found matches
                cosine_similarity = CosineGreedy(tolerance=cosine_tol)
                cosine_scores = []
                for match_id in library_ids[all_match_ids]:
                    cosine_scores.append(cosine_similarity.pair(documents_library[match_id]._obj,
                                                                documents_query[i]._obj))
            else:
                cosine_scores = len(all_match_ids) * ["not calculated"]

            if m_modcos_similarities is not None:
                mod_cosine_scores0 = [x for x in m_modcos_similarities[all_match_ids, i]]
                mod_cosine_scores1 = [x for x in m_modcos_matches[all_match_ids, i]]
                mod_cosine_scores = list(zip(mod_cosine_scores0, mod_cosine_scores1))
            elif "modcosine" in include_scores:
                # Get modified cosine score for found matches
                mod_cosine_similarity = ModifiedCosine(tolerance=cosine_tol)
                mod_cosine_scores = []
                for match_id in library_ids[all_match_ids]:
                    mod_cosine_scores.append(mod_cosine_similarity.pair(documents_library[match_id]._obj,
                                                                        documents_query[i]._obj))
            else:
                mod_cosine_scores = len(all_match_ids) * ["not calculated"]

            matches_df = pd.DataFrame({"cosine_score": [x["score"] for x in cosine_scores],
                                       "cosine_matches": [x["matches"] for x in cosine_scores],
                                       "mod_cosine_score": [x["score"] for x in mod_cosine_scores],
                                       "mod_cosine_matches": [x["matches"] for x in mod_cosine_scores]},
                                       index=library_ids[all_match_ids])

            if m_mass_matches is not None:
                matches_df["mass_match"] = m_mass_matches[all_match_ids, i]

            if m_spec2vec_similarities is not None:
                matches_df["s2v_score"] = m_spec2vec_similarities[all_match_ids, i]
            elif "spec2vec" in include_scores:
                spec2vec_similarity = Spec2Vec(model=model,
                                               intensity_weighting_power=intensity_weighting_power,
                                               allowed_missing_percentage=allowed_missing_percentage)
                spec2vec_scores = []
                for match_id in library_ids[all_match_ids]:
                    spec2vec_scores.append(spec2vec_similarity.pair(documents_library[match_id],
                                                                    documents_query[i]))
                matches_df["s2v_score"] = spec2vec_scores
            found_matches.append(matches_df.fillna(0))
        else:
            found_matches.append([])

    return found_matches



import logging
from .__version__ import __version__


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Netherlands eScience Center"
__email__ = 'f.huber@esciencecenter.nl'
__all__ = [
    "__version__"
]


#!/usr/bin/env python
import os
from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(here, "matchmsextras", "__version__.py")) as f:
    exec(f.read(), version)

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    name="matchmsextras",
    version=version["__version__"],
    description="Additional functionalities to be used with matchms",
    long_description=readme,
    author="Netherlands eScience Center",
    author_email="f.huber@esciencecenter.nl",
    url="https://github.com/matchms/matchmsextras",
    packages=find_packages(exclude=['*tests*']),
    include_package_data=True,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords=[
        "similarity measures",
        "mass spectrometry",
        "networking",
        "network analysis",
        "library search",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    test_suite="tests",
    python_requires='>=3.7',
    install_requires=[
        "matchms",
        "spec2vec",
        "python-louvain",
        "networkx",
        "pandas",
    ],
    extras_require={"dev": ["bump2version",
                            "isort>=4.2.5,<5",
                            "prospector[with_pyroma]",
                            "pytest",
                            "pytest-cov",
                            "sphinx>=3.0.0,!=3.2.0,!=3.5.0,<4.0.0",
                            "sphinx_rtd_theme",
                            "sphinxcontrib-apidoc",
                            "yapf",],
                    "chemistry": ["rdkit >=2020.03.1"]},
)



"""
Functions to create a graph/network from spectrum similarity scores
"""
from typing import List, Tuple
import numpy as np
import networkx as nx
from community import community_louvain
from networkx.algorithms.connectivity import minimum_st_edge_cut  # , minimum_st_node_cut
from networkx.algorithms.flow import shortest_augmenting_path
import pandas as pd
from matplotlib import pyplot as plt
import matplotlib
from matchms import Spectrum, Scores

# ----------------------------------------------------------------------------
# ---------------- Graph / networking related functions ----------------------
# ----------------------------------------------------------------------------

def get_top_hits(scores, identifier: str = "spectrumid",
                 top_n: int = 25, search_by: str = "queries") -> Tuple[dict, dict]:
    """Get top_n highest scores (and indices) for every entry.

    Args:
    --------
    scores
        Matchms Scores object containing all Spec2Vec similarities.
    identifier
        Unique intentifier for each spectrum in scores. Will also be used for
        node names.
    top_n
        Return the indexes and scores for the top_n highest scores.
    search_by
        Chose between 'queries' or 'references' which decides if the top_n matches
        for every spectrum in scores.queries or in scores.references will be
        collected and returned
    """
    assert search_by in ["queries", "references"], \
        "search_by must be 'queries' or 'references"
    if top_n < 2:
        top_n = 2
        print("Set top_n to minimum value of 2")
    #dim1 = len(scores.references) if search_by=="references" else len(scores.queries)
    #dim2 = min(top_n, len(scores.queries) if search_by=="references" else len(scores.references))

    similars_idx = dict()
    similars_scores = dict()

    if search_by=="queries":
        for i, spec in enumerate(scores.queries):
            spec_id = spec.get(identifier)
            similars_idx[spec_id] = scores.scores[:, i].argsort()[::-1][:top_n]
            similars_scores[spec_id] = scores.scores[similars_idx[spec_id], i]
    elif search_by=="references":
        for i, spec in enumerate(scores.references):
            spec_id = spec.get(identifier)
            similars_idx[spec_id] = scores.scores[i, :].argsort()[::-1][:top_n]
            similars_scores[spec_id] = scores.scores[i, similars_idx[spec_id]]
    return similars_idx, similars_scores


def create_network(scores: Scores,
                   identifier: str = "spectrumid",
                   top_n: int = 20,
                   max_links: int = 10,
                   cutoff: float = 0.7,
                   link_method: str = 'single') -> nx.Graph:
    """
    Function to create network from given top-n similarity values. Expects that
    similarities given in scores are from an all-vs-all comparison including all
    possible pairs.

    Args:
    --------
    scores
        Matchms Scores object containing all Spec2Vec similarities.
    identifier
        Unique intentifier for each spectrum in scores. Will also be used for
        node names.
    top_n
        Consider edge between spectrumA and spectrumB if score falls into
        top_n for spectrumA or spectrumB (link_method="single"), or into
        top_n for spectrumA and spectrumB (link_method="mutual"). From those
        potential links, only max_links will be kept, so top_n must be >= max_links.
    max_links
        Maximum number of links to add per node. Default = 10.
        Due to incoming links, total number of links per node can be higher.
    cutoff
        Threshold for given similarities. Edges/Links will only be made for
        similarities > cutoff. Default = 0.7.
    link_method
        Chose between 'single' and 'mutual'. 'single will add all links based
        on individual nodes. 'mutual' will only add links if that link appears
        in the given top-n list for both nodes.
    """
    assert top_n >= max_links, "top_n must be >= max_links"
    assert np.all(scores.queries == scores.references), \
        "Expected symmetric scores object with queries==references"
    unique_ids = list({s.get(identifier) for s in scores.queries})
    dimension = len(unique_ids)

    # Initialize network graph, add nodes
    msnet = nx.Graph()
    msnet.add_nodes_from(unique_ids)

    # Collect location and score of highest scoring candidates for queries and references
    similars_idx, similars_scores = get_top_hits(scores, top_n=top_n,
                                                 search_by="queries")

    # Add edges based on global threshold (cutoff) for weights
    for i, spec in enumerate(scores.queries):
        query_id = spec.get(identifier)

        ref_candidates = np.array([scores.references[x].get(identifier)
                                   for x in similars_idx[query_id]])
        idx = np.where((similars_scores[query_id] >= cutoff) & (ref_candidates != query_id))[0][:max_links]
        if link_method == "single":
            new_edges = [(query_id, str(ref_candidates[x]),
                          float(similars_scores[query_id][x])) for x in idx]
        elif link_method == "mutual":
            new_edges = [(query_id, str(ref_candidates[x]),
                          float(similars_scores[query_id][x]))
                         for x in idx if i in similars_idx[ref_candidates[x]][:]]
        else:
            raise ValueError("Link method not kown")

        msnet.add_weighted_edges_from(new_edges)

    return msnet


def create_network_asymmetric(scores: Scores,
                              identifier: str = "spectrumid",
                              add_links_from_queries: bool = True,
                              add_links_from_references: bool = False,
                              top_n: int = 20,
                              max_links: int = 10,
                              cutoff: float = 0.7,
                              link_method: str = 'single') -> nx.Graph:
    """
    Function to create network from given top-n similarity values. Expects scores
    object where queries != references.

    Args:
    --------
    scores
        Matchms Scores object containing all Spec2Vec similarities.
    identifier
        Unique intentifier for each spectrum in scores. Will also be used for
        node names.
    top_n
        Consider edge between spectrumA and spectrumB if score falls into
        top_n for spectrumA or spectrumB (link_method="single"), or into
        top_n for spectrumA and spectrumB (link_method="mutual"). From those
        potential links, only max_links will be kept, so top_n must be >= max_links.
    max_links
        Maximum number of links to add per node. Default = 10.
        Due to incoming links, total number of links per node can be higher.
    cutoff
        Threshold for given similarities. Edges/Links will only be made for
        similarities > cutoff. Default = 0.7.
    link_method
        Chose between 'single' and 'mutual'. 'single will add all links based
        on individual nodes. 'mutual' will only add links if that link appears
        in the given top-n list for both nodes.
    """
    assert top_n >= max_links, "top_n must be >= max_links"
    set_q = {s.get(identifier) for s in scores.queries}
    set_r = {s.get(identifier) for s in scores.references}
    unique_ids = list(set_q.union(set_r))
    dimension = len(unique_ids)

    # Initialize network graph, add nodes
    msnet = nx.Graph()
    msnet.add_nodes_from(unique_ids)

    # Collect location and score of highest scoring candidates for queries and references
    similars_idx_q, similars_scores_q = get_top_hits(scores, top_n=top_n,
                                                     search_by="queries")
    similars_idx_r, similars_scores_r = get_top_hits(scores, top_n=top_n,
                                                     search_by="references")
    if add_links_from_queries:
        # Add edges based on global threshold (cutoff) for weights
        for i, spec in enumerate(scores.queries):
            query_id = spec.get(identifier)

            ref_candidates = np.array([scores.references[x].get(identifier)
                                       for x in similars_idx_q[query_id]])
            idx = np.where((similars_scores_q[query_id] >= cutoff) & (ref_candidates != query_id))[0][:max_links]
            if link_method == "single":
                new_edges = [(query_id, ref_candidates[x],
                              float(similars_scores_q[query_id][x])) for x in idx]
            elif link_method == "mutual":
                new_edges = [(query_id, ref_candidates[x],
                              float(similars_scores_q[query_id][x]))
                             for x in idx if i in similars_idx_r[ref_candidates[x]][:]]
            else:
                raise ValueError("Link method not kown")

            msnet.add_weighted_edges_from(new_edges)

    if add_links_from_references:
        # Add edges based on global threshold (cutoff) for weights
        for i, spec in enumerate(scores.references):
            ref_id = spec.get(identifier)

            query_candidates = np.array([scores.queries[x].get(identifier)
                                         for x in similars_idx_r[ref_id]])
            idx = np.where((similars_scores_r[ref_id] >= cutoff) & (query_candidates != ref_id))[0][:max_links]
            if link_method == "single":
                new_edges = [(ref_id, query_candidates[x],
                              float(similars_scores_r[ref_id][x])) for x in idx]
            elif link_method == "mutual":
                new_edges = [(ref_id, query_candidates[x],
                              float(similars_scores_r[ref_id][x])) for x in idx
                             if i in similars_idx_q[query_candidates[x]][:]]
            else:
                raise ValueError("Link method not kown")

            msnet.add_weighted_edges_from(new_edges)

    return msnet


def extract_networking_metadata(spectrums: List[Spectrum],
                                identifier: str = "spectrumid") -> pd.DataFrame:
    """Collect metadata to later visualize and play with the network.

    Parameters
    ----------
    spectrums
        List of spectrums

    """
    identifiers = []
    smiles = []
    compound_names = []
    parent_masses = []
    for spec in spectrums:
        identifiers.append(spec.get(identifier))
        smiles.append(spec.get("smiles"))
        compound_names.append(spec.get("compound_name"))
        parent_masses.append(spec.get("parent_mass"))

    metadata = pd.DataFrame({'smiles': smiles,
                             'compound_name': compound_names,
                             'parent_mass': parent_masses} )
    metadata.index = identifiers
    return metadata


def sample_cuts(graph, max_steps=1000, max_cuts=1):
    """ Function to help find critical links in the given graph.
    Critical links here are links which -once removed- would disconnect considerable
    parts of the network. Those links are searched for by counting minimum cuts between
    a large number of node pairs (up to max_steps pairs will be explored).
    If more pairs exist than max_steps allows to explore, pick max_steps random pairs.

    Args:
    -------
    graph: networkx graph
        Graph of individual cluster (created using networkx).
    max_steps
        Up to max_steps pairs will be explored to search for cuts. Default = 1000.
    max_cuts
        Maximum numbers of links allowed to be cut. Default = 1.
    """

    num_nodes = graph.number_of_nodes()
    # num_edges = graph.number_of_edges()

    # Make list of all pairs within graph
    nodes = np.array(graph.nodes)
    pairs = np.array(np.meshgrid(nodes, nodes)).T
    remove_diagonal = np.array([(i * num_nodes + i) for i in range(num_nodes)])
    pairs = np.delete(pairs.reshape(-1, 2), remove_diagonal, axis=0)

    sampled_cuts = []
    if pairs.shape[0] <= max_steps:
        max_steps = pairs.shape[0]
    else:
        # If more pairs exist than max_steps allows to explore, pick max_steps random pairs.
        choices = np.random.choice(np.arange(pairs.shape[0]),
                                   max_steps,
                                   replace=False)
        pairs = pairs[choices, :]

    for pair in pairs:
        cuts = minimum_st_edge_cut(graph,
                                   pair[0],
                                   pair[1],
                                   flow_func=shortest_augmenting_path)
        # nx.node_connectivity(graphs[4], 592, 376)
        # cuts = nx.minimum_st_edge_cut(graph, pair[0], pair[1])
        # cuts = nx.minimum_edge_cut(graph, pair[0], pair[1])#, flow_func=shortest_augmenting_path)
        if len(cuts) <= max_cuts:
            sampled_cuts.append(cuts)

    return sampled_cuts


def weak_link_finder(graph, max_steps=1000, max_cuts=1):
    """ Function to detect critical links in the given graph.
    Critical links here are links which -once removed- would disconnect considerable
    parts of the network. Those links are searched for by counting minimum cuts between
    a large number of node pairs (up to max_steps pairs will be explored).
    If more pairs exist than max_steps allows to explore, pick max_steps random pairs.

    Args:
    -------
    graph: networkx graph
        Graph of individual cluster (created using networkx).
    max_steps
        Up to max_steps pairs will be explored to search for cuts. Default = 1000.
    max_cuts
        Maximum numbers of links allowed to be cut. Default = 1.
    """

    sampled_cuts = sample_cuts(graph, max_steps=max_steps, max_cuts=max_cuts)

    sampled_cuts_len = [len(x) for x in sampled_cuts]
    proposed_cuts = []
    for min_cuts in list(set(sampled_cuts_len)):
        sampled_cuts_select = [
            list(x)[:min_cuts] for x in sampled_cuts if len(x) == min_cuts
        ]

        sampled_cuts_select = np.array(sampled_cuts_select)
        # Sort array
        if min_cuts > 1:
            sampled_cuts_select = np.sort(np.sort(sampled_cuts_select, axis=2),
                                          axis=1)
        else:
            sampled_cuts_select = np.sort(sampled_cuts_select, axis=2)

        # Find unique cuts and count occurences
        cuts_unique, cuts_count = row_counts(
            sampled_cuts_select.reshape(-1, min_cuts * 2))

        # Return most promising cuts
        proposed_cuts.append((min_cuts, cuts_unique, cuts_count))

    return proposed_cuts


def dilate_cluster(graph_main: nx.Graph,
                   scores: Scores,
                   identifier: str = "spectrumid",
                   max_cluster_size: int = 100,
                   min_cluster_size: int = 10,
                   max_per_node: int = 1,
                   max_per_cluster=None,
                   min_weight: float = 0.5,
                   top_n: int = 25):
    """ Add more links to clusters that are < min_cluster_size.
    This function is in particular made to avoid small remaining clusters or singletons.

    Will only add links if they won't lead to clusters > max_cluster_size,
    and if the links have weights > min_weight.
    Starts iteratively from highest weight links that are not yet part of the network
    (out of given top-n links).

    Args:
    --------
    graph_main
        Graph, e.g. made using create_network() function. Based on networkx.
    scores
        Matchms Scores object containing all Spec2Vec similarities.
    identifier
        Unique intentifier for each spectrum in scores. Will also be used for
        node names.
    max_cluster_size
        Maximum desired size of clusters. Default = 100.
    min_cluster_size
        Minimum desired size of clusters. Default = 10.
    max_per_node
        Only add the top max_addition ones per cluster. Default = 1.
    max_per_cluster: int, None
        Only add the top max_addition ones per cluster. Ignore if set to None. Default = None.
    min_weight
        Set minimum weight to be considered for making link. Default = 0.5.
    top_n
        Consider the top_n highest scores for addition during dilation step. Default is 25.
    """

    links_added = []

    similars_idx, similars_scores = get_top_hits(scores, top_n=top_n,
                                                 search_by="queries")

    # Split graph into separate clusters
    graphs = [graph_main.subgraph(c).copy() for c in nx.connected_components(graph_main)]
    #graphs = list(nx.connected_component_subgraphs(graph_main))

    for graph in graphs:
        cluster_size = len(graph.nodes)
        if cluster_size < min_cluster_size:
            best_scores = []
            potential_links = []

            for ID in graph.nodes:
                nodes_connected = []
                for key in graph[ID].keys():
                    nodes_connected.append(key)
                similars_identifiers = [scores.queries[x].get(identifier) for x in similars_idx[ID]]
                potential_new_links = [(i, x)
                                       for i, x in enumerate(similars_identifiers)
                                       if x not in nodes_connected and x != ID]
                best_score_arr = similars_scores[ID][[
                    x[0] for x in potential_new_links
                ]]
                select = np.where(
                    best_score_arr >= min_weight)[0][:max_per_node]
                # if best_score >= min_weight:
                if select.shape[0] > 0:
                    for s in select:
                        best_scores.append(best_score_arr[s])
                        potential_link = (ID,
                                          [x[1]
                                           for x in potential_new_links][s])
                        potential_links.append(potential_link)

            if max_per_cluster is None:
                selected_candidates = np.argsort(best_scores)[::-1]
            else:
                # Only add the top max_addition ones
                selected_candidates = np.argsort(
                    best_scores)[::-1][:max_per_cluster]

            for ID in selected_candidates:
                # node_id = list(graph.nodes)[ID]
                node_id = potential_links[ID][0]

                # Only add link if no cluster > max_cluster_size is formed by it
                if (len(
                        nx.node_connected_component(graph_main,
                                                    potential_links[ID][1])) +
                        cluster_size) <= max_cluster_size:
                    # Actual adding of new links
                    graph_main.add_edge(node_id,
                                        potential_links[ID][1],
                                        weight=best_scores[ID])
                    links_added.append((node_id, potential_links[ID][1]))
                    # Update cluster_size to keep track of growing clusters
                    cluster_size = len(
                        nx.node_connected_component(graph_main,
                                                    potential_links[ID][1]))

    return graph_main, links_added


def erode_clusters(graph_main, max_cluster_size=100, keep_weights_above=0.8):
    """ Remove links from clusters that are > max_cluster_size.
    This function is in particular made to avoid small remaining clusters or singletons.

    Will only add links if they won't lead to clusters > max_cluster_size,
    and if the links have weights > min_weight.
    Starts iteratively from highest weight links that are not yet part of the network.

    Args:
    --------
    graph_main: networkx graph
        Graph, e.g. made using create_network() function. Based on networkx.
    max_cluster_size: int
        Maximum desired size of clusters. Default = 100.
    keep_weights_above: float
        Set threshold above which weights will not be removed. Default = 0.8.
    """

    links_removed = []

    # Split graph into separate clusters
    graphs = list(nx.connected_component_subgraphs(graph_main))

    for graph in graphs:
        cluster_size = len(graph.nodes)
        while cluster_size > max_cluster_size:

            edges = list(graph.edges)
            edges_weights = np.array(
                [graph[x[0]][x[1]]['weight'] for x in edges])

            weakest_edge = edges_weights.argsort()[0]
            if edges_weights[weakest_edge] < keep_weights_above:
                print("Remove edge:", edges[weakest_edge][0],
                      edges[weakest_edge][1])
                graph.remove_edge(edges[weakest_edge][0],
                                  edges[weakest_edge][1])
                graph_main.remove_edge(edges[weakest_edge][0],
                                       edges[weakest_edge][1])
                links_removed.append(edges[weakest_edge])

            # If link removal caused split of cluster:
            if not nx.is_connected(graph):
                subgraphs = list(nx.connected_component_subgraphs(graph))
                print("Getting from cluster with", len(graph.nodes),
                      "nodes, to clusters with",
                      [len(x.nodes) for x in subgraphs], "nodes.")
                idx1 = np.argmax([len(x.nodes) for x in subgraphs])
                graph = subgraphs[idx1]  # keep largest subcluster here

            cluster_size = len(graph.nodes)

    return graph_main, links_removed


def add_intra_cluster_links(graph_main, scores, identifier="spectrumid", min_weight=0.5, max_links=20):
    """ Add links within each separate cluster if weights above min_weight.

    Args:
    -------
    graph_main: networkx graph
        Graph, e.g. made using create_network() function. Based on networkx.
    scores: matchms Scores
        2D array with all reference similarity values between all-vs-all nodes.
    min_weight: float
        Set minimum weight to be considered for making link. Default = 0.5.
    """
    # Split graph into separate clusters
    graphs = [graph_main.subgraph(c).copy() for c in nx.connected_components(graph_main)]
    node_to_ID = {s.get(identifier): i for i, s in enumerate(scores.queries)}
    for graph in graphs:
        nodes = list(graph.nodes)
        nodes0 = nodes.copy()
        for node in nodes:
            del nodes0[0]
            ID1 = node_to_ID[node]
            IDs2 = [node_to_ID[n] for n in nodes0]
            weights = scores.scores[ID1, IDs2]
            weights_select = weights.argsort()[::-1][:max_links]
            weights_select = np.where(weights[weights_select] >= min_weight)[0]
            new_edges = [(node, nodes0[x], weights[x]) for x in weights_select]

            graph_main.add_weighted_edges_from(new_edges)

    return graph_main


def split_cluster(graph_main,
                  max_cluster_size=100,
                  min_cluster_size=10,
                  max_search_steps=1000,
                  max_cuts=1,
                  multiple_cuts_per_level=True):
    """
    Function to split clusters at weak links.

    Args:
    ---------
    graph_main: networkx graph
        Graph, e.g. made using create_network() function. Based on networkx.
    max_cluster_size: int
        Maximum desired size of clusters. Default = 100.
    min_cluster_size: int
        Minimum desired size of clusters. Default = 10.
    max_steps
        Up to max_steps pairs will be explored to search for cuts. Default = 1000.
    max_cuts
        Maximum numbers of links allowed to be cut. Default = 1.
    multiple_cuts_per_level
        If true allow multiple cuts to be done per level and run. Default = True.
    """

    # Split graph into separate clusters
    graphs = list(nx.connected_component_subgraphs(graph_main))

    links_removed = []
    for i, graph in enumerate(graphs):
        if len(graph.nodes) > max_cluster_size:
            # Detect potential weak links
            weak_links = weak_link_finder(graph,
                                          max_steps=max_search_steps,
                                          max_cuts=max_cuts)

            split_done = False
            j = 0
            new_graph = graph.copy()
            while not split_done and j < len(weak_links):

                # Test best candidates

                new_graph_testing = new_graph.copy()
                pairs = weak_links[j][1]
                pair_counts = weak_links[j][2]
                pairs = pairs[pair_counts.argsort()[::-1]]
                # print(i,j, pairs)

                # ----------------------------------------------
                # Check if pairs have already been removed in former iteration
                # ----------------------------------------------
                pairs_still_present = []
                for i, pair in enumerate(pairs):
                    all_edges_present = True
                    for m in range(int(pairs.shape[1] / 2)):
                        edge = (pair[m * 2], pair[m * 2 + 1])
                        if edge not in new_graph_testing.edges:
                            all_edges_present = False
                    if all_edges_present:
                        pairs_still_present.append(i)
                    pairs_still_present = list(set(pairs_still_present))
                pairs = pairs[
                    pairs_still_present]  # Remove pairs which have been cut out already

                # ----------------------------------------------
                # Test removing proposed links for all pairs
                # ----------------------------------------------
                if len(pairs) > 0:
                    min_size_after_cutting = []
                    for pair in pairs:
                        new_graph_testing = new_graph.copy()

                        # Remove edges in pair
                        for m in range(int(pairs.shape[1] / 2)):
                            new_graph_testing.remove_edge(
                                pair[m * 2], pair[m * 2 + 1])

                        # Check if created subclustes are big enough:
                        subgraphs = list(
                            nx.connected_component_subgraphs(
                                new_graph_testing))
                        min_size_after_cutting.append(
                            min([len(x.nodes) for x in subgraphs]))

                    # Select best partition of graph (creating most similar sized subclusters)
                    min_size_after_cutting = np.array(min_size_after_cutting)
                    best_partition = np.argmax(min_size_after_cutting)
                else:
                    min_size_after_cutting = [0]
                    best_partition = 0

                # ----------------------------------------------
                # Actual removal of links
                # ----------------------------------------------
                if min_size_after_cutting[best_partition] >= min_cluster_size:
                    new_graph_testing = new_graph.copy()
                    pair = pairs[best_partition]

                    # Remove edges in selected pair
                    for m in range(int(pairs.shape[1] / 2)):
                        # Remove edge from current cluster:
                        new_graph_testing.remove_edge(pair[m * 2],
                                                      pair[m * 2 + 1])
                        # Remove edge from main graph:
                        graph_main.remove_edge(pair[m * 2], pair[m * 2 + 1])
                        links_removed.append((pair[m * 2], pair[m * 2 + 1]))
                    subgraphs = list(
                        nx.connected_component_subgraphs(new_graph_testing))

                    if int(pairs.shape[1] / 2) > 1:
                        print("Removed", int(pairs.shape[1] / 2), "edges:",
                              pair)
                    else:
                        print("Removed", int(pairs.shape[1] / 2), "edge:",
                              pair)

                    print("Getting from cluster with", len(new_graph.nodes),
                          "nodes, to clusters with",
                          [len(x.nodes) for x in subgraphs], "nodes.")
                    idx1 = np.argmax([len(x.nodes) for x in subgraphs])
                    new_graph = subgraphs[idx1]  # keep largest subcluster here

                    if len(new_graph.nodes) <= max_cluster_size:
                        split_done = True
                    else:
                        pass

                # Check if more suited cuts are expected for the same number of cuts
                if len(min_size_after_cutting) > 1:
                    idx = np.argsort(min_size_after_cutting)[::-1][1]
                    if min_size_after_cutting[
                            idx] >= min_cluster_size and multiple_cuts_per_level:
                        pass
                    else:
                        j += 1
                else:
                    j += 1

    return graph_main, links_removed


# ----------------------------------------------------------------------------
# ---------------------- Functions to refine network -------------------------
# ----------------------------------------------------------------------------


def refine_network(graph_main,
                   similars_idx,
                   similars,
                   weigh_bounds=(0.6, 1),
                   filename=None,
                   max_cluster_size=100,
                   min_cluster_size=10,
                   max_search_steps=1000,
                   max_cuts=2,
                   max_split_iterations=10,
                   basic_splitting=True,
                   dilation=False):
    """
    Args:
    -------
    """
    # Split graph into separate clusters
    graphs = list(nx.connected_component_subgraphs(graph_main))

    links_removed = []
    links_added = []

    # n_cluster = len(graphs)
    cluster_max = np.max([len(x.nodes) for x in graphs])
    counter = 0

    print(20 * '---')
    while cluster_max > max_cluster_size and counter < max_split_iterations:
        print("Splitting iteration:", counter + 1, "Max cluster size =",
              cluster_max, '\n')
        graph_main, links = split_cluster(graph_main.copy(),
                                          max_cluster_size=max_cluster_size,
                                          min_cluster_size=min_cluster_size,
                                          max_search_steps=max_search_steps,
                                          max_cuts=max_cuts,
                                          multiple_cuts_per_level=True)
        links_removed.extend(links)

        # Split updated graph into separate clusters
        graphs = list(nx.connected_component_subgraphs(graph_main))
        cluster_max = np.max([len(x.nodes) for x in graphs])
        counter += 1

    if basic_splitting:
        print(20 * '---')
        print("Extra splitting step to sanitize clusters.")
        graph_main, links = split_cluster(
            graph_main,
            max_cluster_size=2 *
            min_cluster_size,  # ! here we try to 'sanitize most clusters'
            min_cluster_size=min_cluster_size,
            max_search_steps=max_search_steps,
            max_cuts=1,
            multiple_cuts_per_level=False)
        links_removed.extend(links)

    if dilation:
        print(20 * '---')
        print("Runing dilation function for smaller clusters <",
              min_cluster_size)
        graph_main, links = dilate_cluster(graph_main,
                                           similars_idx,
                                           similars,
                                           max_cluster_size=max_cluster_size,
                                           min_cluster_size=min_cluster_size,
                                           min_weight=weigh_bounds[0])
        links_added.extend(links)

    if filename is not None:
        # Export graph for drawing (e.g. using Cytoscape)
        nx.write_graphml(graph_main, filename)
        print("Network stored as graphml file under: ", filename)

    return graph_main, links_added, links_removed


# ----------------------------------------------------------------------------
# -------------------- Functions to evaluate networks ------------------------
# ----------------------------------------------------------------------------


def evaluate_clusters(graph_main, m_sim_ref):
    """ Evaluate separate clusters of network based on given reference matrix.

    Args:
    -------
    graph_main: networkx graph
        Graph, e.g. made using create_network() function. Based on networkx.
    m_sim_ref: numpy array
        2D array with all reference similarity values between all-vs-all nodes.
    """

    # Split graph into separate clusters
    graphs = list(nx.connected_component_subgraphs(graph_main))

    num_nodes = []
    num_edges = []
    ref_sim_mean_edges = []
    ref_sim_var_edges = []
    ref_sim_mean_nodes = []
    ref_sim_var_nodes = []

    # Loop through clusters
    for graph in graphs:
        num_nodes.append(len(graph.nodes))
        if len(graph.edges) > 0:  # no edges for singletons
            num_edges.append(len(graph.edges))

            edges = list(graph.edges)
            mol_sim_edges = np.array([m_sim_ref[x] for x in edges])
            mol_sim_edges = np.nan_to_num(mol_sim_edges)
            ref_sim_mean_edges.append(np.mean(mol_sim_edges))
            ref_sim_var_edges.append(np.var(mol_sim_edges))
        else:
            num_edges.append(0)
            ref_sim_mean_edges.append(0)
            ref_sim_var_edges.append(0)

        nodes = list(graph.nodes)
        mean_mol_sims = []
        for node in nodes:
            mean_mol_sims.append(m_sim_ref[node, nodes])

        ref_sim_mean_nodes.append(np.mean(mean_mol_sims))
        ref_sim_var_nodes.append(np.var(mean_mol_sims))

    zipped = zip(num_nodes, num_edges, ref_sim_mean_edges, ref_sim_var_edges, ref_sim_mean_nodes, ref_sim_var_nodes)
    cluster_data = pd.DataFrame(list(zipped), columns=['num_nodes', 'num_edges', 'ref_sim_mean_edges',
                                                       'ref_sim_var_edges', 'ref_sim_mean_nodes', 'ref_sim_var_nodes'])
    return cluster_data


def evaluate_clusters_louvain(graph_main: nx.Graph,
                              reference_scores: pd.DataFrame, resolution: float = 1.0):
    """ Cluster given network using Louvain algorithm.
    Then evaluate clusters of network based on given reference matrix.

    Args:
    -------
    graph_main
        Graph, e.g. made using create_network() function. Based on networkx.
    reference_scores
        Pandas DataFrame with all reference similarity values between all-vs-all nodes.
        Expects index and column values to be inchikeys14.
    resolution
        Louvain algorithm resolution parameter. Will change size of communities.
        See also: https://python-louvain.readthedocs.io/en/latest/api.html Default=1.0
    """
    # Find clusters using Louvain algorithm (and python-louvain library)
    communities = community_louvain.best_partition(graph_main,
                                                   weight='weight',
                                                   resolution=resolution)
    nx.set_node_attributes(graph_main, communities, 'modularity')

    clusters = []
    for cluster_id in set(communities.values()):
        cluster = [
            nodes for nodes in communities.keys()
            if communities[nodes] == cluster_id
        ]
        clusters.append(cluster)

    num_nodes = []
    ref_sim_mean_nodes = []
    ref_sim_var_nodes = []

    for cluster in clusters:
        num_nodes.append(len(cluster))
        mean_mol_sims = []
        # TODO: switch from new ids to location indices (node, cluster)
        for node in cluster:
            mean_mol_sims.append(m_sim_ref[node, cluster])

        ref_sim_mean_nodes.append(np.mean(mean_mol_sims))
        ref_sim_var_nodes.append(np.var(mean_mol_sims))

    cluster_data = pd.DataFrame(
        list(zip(num_nodes, ref_sim_mean_nodes, ref_sim_var_nodes)),
        columns=['num_nodes', 'ref_sim_mean_nodes', 'ref_sim_var_nodes'])

    return graph_main, cluster_data


# ----------------------------------------------------------------------------
# --------------------- Graph related plotting functions ---------------------
# ----------------------------------------------------------------------------
def plots_cluster_evaluations(cluster_data_collection,
                              total_num_nodes,
                              size_bins,
                              labels,
                              title,
                              filename=None):
    """ Plot cluster sizes and mean node similarity.

    Args:
    --------
    cluster_data_collection:  list
        List of cluster data for all scenarios to be plotted.
    total_num_nodes: int
        Total number of nodes of graph.
    size_bins: list of int
        List of bins for cluster sizes.
    labels: list of str
        List of labels for all scenarios in list of cluster_data_collection.
    title: str
        Title for plot. Default = None
    filename: str
        If not None: save figure to file with given name.
    """

    plt.style.use('ggplot')
    fig = plt.figure(figsize=(12, 5))
    ax = plt.subplot(111)

    num_plots = len(cluster_data_collection)
    cmap = matplotlib.cm.get_cmap('inferno')  # 'Spectral')
    bins = [0] + [x + 1 for x in size_bins]
    x_labels = ['<' + str(bins[1])]
    x_labels += [
        str(bins[i]) + '-' + str(bins[i + 1] - 1)
        for i in range(1,
                       len(bins) - 2)
    ]
    x_labels += ['>' + str(bins[-2] - 1)]

    for count, cluster_data in enumerate(cluster_data_collection):

        num_elements = []
        mean_edge_sim = []
        mean_node_sim = []
        for i in range(len(bins) - 1):
            num_elements.append(
                np.sum(cluster_data[(cluster_data['num_nodes'] <= bins[i + 1])
                                    & (cluster_data['num_nodes'] > bins[i])]
                       ['num_nodes'].values))

            if 'ref_sim_mean_edges' in cluster_data.columns:
                mean_edge_sim.append(
                    np.mean(
                        cluster_data[(cluster_data['num_nodes'] <= bins[i + 1])
                                     & (cluster_data['num_nodes'] > bins[i])]
                        ['ref_sim_mean_edges'].values))

            mean_node_sim.append(
                np.mean(cluster_data[(cluster_data['num_nodes'] <= bins[i + 1])
                                     & (cluster_data['num_nodes'] > bins[i])]
                        ['ref_sim_mean_nodes'].values))

        num_elements[0] = total_num_nodes - np.sum(num_elements[1:])
        if 'ref_sim_mean_edges' in cluster_data.columns:
            if np.isnan(mean_edge_sim[0]):
                mean_edge_sim[0] = 0

        plt.scatter(x_labels,
                    mean_node_sim,
                    s=num_elements,
                    facecolor="None",
                    edgecolors=[cmap(count / num_plots)],
                    lw=3,
                    alpha=0.7,
                    label=labels[count])

    plt.xlabel('cluster size')
    plt.ylabel('mean molecular similarity of nodes in cluster')
    chartbox = ax.get_position()
    ax.set_position(
        [chartbox.x0, chartbox.y0, chartbox.width * 0.8, chartbox.height])
    lgnd = ax.legend(loc='upper center', bbox_to_anchor=(1.12, 1))
    for i in range(num_plots):
        lgnd.legendHandles[i]._sizes = [30]

    plt.title(title)

    # Save figure to file
    if filename is not None:
        plt.savefig(filename, dpi=600)


def plot_clustering_performance(data_collection,
                                labels,
                                thres_well=0.5,
                                thres_poor=0.5,
                                title=None,
                                filename=None,
                                size_xy=(8, 5)):
    """ Plot cluster evaluations for all conditions found in data_collection.
    Cluster will be classified as "well clustered" if the mean(similarity) across
    all nodes is > thres_well. Or as "poorly clustered" if < thres_poor.
    Clusters with only one node (singletons) will be counted as "non-clustered".

    Args:
    --------
    data_collection: list of pandas.DataFrame()
        List of DataFrames as created by evaluate_clusters().
    labels: list
        List of labels for the different conditions found in data_collection.
    thres_well: float
        Threshold above which clusters will be classified as "well clustered". Default = 0.5.
    thres_poor: float
        Threshold below which clusters will be classified as "poorly clustered". Default = 0.5.
    title: str
        Title for plot. Default = None
    filename: str
        If not none: save figure to file with given name.
    size_xy: tuple
        Figure size. Default is (8,5).
    """
    plt.style.use('ggplot')
    performance_data = []
    ymax = np.sum(data_collection[0]['num_nodes'].values) # total_num_nodes
    legend_labels = [
        "average structural similarity >= {:.2}".format(thres_well),
        "average structural similarity < {:.2}".format(thres_poor),
        'non-clustered nodes'
    ]

    for cluster_data in data_collection:
        nodes_clustered_well = np.sum(
            cluster_data[(cluster_data['num_nodes'] > 1)
                         & (cluster_data['ref_sim_mean_nodes'] >= thres_well)]
            ['num_nodes'].values)
        nodes_clustered_poor = np.sum(
            cluster_data[(cluster_data['num_nodes'] > 1)
                         & (cluster_data['ref_sim_mean_nodes'] < thres_poor)]
            ['num_nodes'].values)
        nodes_not_clustered = np.sum(
            cluster_data[(cluster_data['num_nodes'] < 2)]['num_nodes'].values)

        performance_data.append(
            [nodes_clustered_well, nodes_clustered_poor, nodes_not_clustered])

    fig = plt.figure(figsize=size_xy)
    ax = plt.subplot(111)
    ax.set_aspect(aspect=5)
    plt.plot(labels, [x[0] / ymax for x in performance_data],
             'o-',
             color='crimson',
             label=legend_labels[0])
    plt.plot(labels, [x[1] / ymax for x in performance_data],
             'o-',
             color='teal',
             label=legend_labels[1])
    plt.plot(labels, [x[2] / ymax for x in performance_data],
             'o-',
             color='darkblue',
             alpha=0.6,
             label=legend_labels[2])
    plt.title(title)
    plt.ylabel("Fraction of total nodes")
    plt.xlabel("networking conditions")
    plt.ylim(-0.03, 1)
    plt.legend()

    # Place legend
    # chartbox = ax.get_position()
    # ax.set_position([chartbox.x0, chartbox.y0, chartbox.width*0.8, chartbox.height])
    # ax.legend(loc='upper center', bbox_to_anchor=(1.25, 1))

    # Save figure to file
    if filename is not None:
        plt.savefig(filename, dpi=600)


def plot_cluster(g, filename=None):
    """ Very basic plotting function to inspect small to medium sized clusters (or networks).

    Args:
    --------
    g: networkx.Graph
        Networkx generated graph containing nodes and edges.
    filename: str
        If not none: save figure to file with given name.
    """
    if len(g.nodes) > 1:
        edges = [(u, v) for (u, v, d) in g.edges(data=True)]
        weights = [d['weight'] for (u, v, d) in g.edges(data=True)]
        weights = weights - 0.95 * np.min(weights)
        weights = weights / np.max(weights)

        # Positions for all nodes
        pos = nx.spring_layout(g)

        plt.figure(figsize=(12, 12))

        # Nodes
        nx.draw_networkx_nodes(g, pos, node_size=100)

        # Edges
        nx.draw_networkx_edges(g,
                               pos,
                               edgelist=edges,
                               width=4 * weights,
                               alpha=0.5)

        # Labels
        nx.draw_networkx_labels(g, pos, font_size=5, font_family='sans-serif')

        plt.axis('off')
        plt.show()

        if filename is not None:
            plt.savefig(filename, dpi=600)
    else:
        print("Given graph has not enough nodes to plot network.")


# ----------------------------------------------------------------------------
# -------------------------- Small helper functions --------------------------
# ----------------------------------------------------------------------------


def row_counts(array):
    """
    Function to find unique rows and count their occurences.
    """
    dt = np.dtype((np.void, array.dtype.itemsize * array.shape[1]))
    b = np.ascontiguousarray(array).view(dt)
    unq, cnt = np.unique(b, return_counts=True)
    unq = unq.view(array.dtype).reshape(-1, array.shape[1])

    return unq, cnt
